@@ -4,24 +4,25 @@ import json
 import operator as op
 import re
 import sys
-from collections import Counter, defaultdict
+from collections import Counter
 from datetime import date, datetime
 from functools import partial
-from html import unescape
 from typing import Any, Dict, Iterable, List, Optional, Set
 from unicodedata import normalize
 
 from beets import config as beets_config
 from beets.autotag.hooks import AlbumInfo, TrackInfo
+from ordered_set import OrderedSet as ordset  # type: ignore[import]
 from pkg_resources import get_distribution, parse_version
 from pycountry import countries, subdivisions
 
 from ._helpers import PATTERNS, Helpers, MediaInfo
+from ._tracks import Track, Tracks
 
 if sys.version_info.minor > 7:
     from functools import cached_property  # pylint: disable=ungrouped-imports
 else:
-    from cached_property import cached_property  # type: ignore
+    from cached_property import cached_property  # type: ignore # pylint: disable=import-error # noqa
 
 NEW_BEETS = get_distribution("beets").parsed_version >= parse_version("1.5.0")
 
@@ -40,12 +41,6 @@ DIGI_MEDIA = "Digital Media"
 VA = "Various Artists"
 
 
-def urlify(pretty_string: str) -> str:
-    """Transform a string into bandcamp url."""
-    name = pretty_string.lower().replace("'", "").replace(".", "")
-    return re.sub("--+", "-", re.sub(r"\W", "-", name, flags=re.ASCII)).strip("-")
-
-
 class Metaguru(Helpers):
     _singleton = False
     va_name = VA
@@ -54,6 +49,7 @@ class Metaguru(Helpers):
     meta: JSONDict
     config: JSONDict
     media_formats: List[MediaInfo]
+    _tracks: Tracks
 
     def __init__(self, meta: JSONDict, config: Optional[JSONDict] = None) -> None:
         self.meta = meta
@@ -64,16 +60,16 @@ class Metaguru(Helpers):
             self.media = self.media_formats[0]
         self.config = config or {}
         self.va_name = beets_config["va_name"].as_str() or self.va_name
+        self._tracks = Tracks.from_json(meta)
 
     @classmethod
     def from_html(cls, html: str, config: JSONDict = None) -> "Metaguru":
         try:
-            meta = json.loads(re.search(PATTERNS["meta"], html).group())  # type: ignore
-            meta["tracks"] = list(map(unescape, re.findall(r"^[0-9]+[.] .*", html, re.M)))
+            meta = re.search(PATTERNS["meta"], html).group()  # type: ignore[union-attr]
         except AttributeError as exc:
             raise AttributeError("Could not find release metadata JSON") from exc
-
-        return cls(meta, config)
+        else:
+            return cls(json.loads(meta), config)
 
     @cached_property
     def excluded_fields(self) -> Set[str]:
@@ -101,19 +97,43 @@ class Metaguru(Helpers):
         )
 
     @cached_property
-    def original_album_name(self) -> str:
-        return self.meta["name"]
+    def official_album_name(self) -> str:
+        """Check description for the album name header and return whatever follows it
+        if found.
+        """
+        m = re.search(r"(Title: ?|Album(:|/Single) )([^\n]+)", self.all_media_comments)
+        if m:
+            return m.group(3).strip()
+        return ""
 
     @cached_property
     def parsed_album_name(self) -> str:
-        match = re.search(
-            r"(Title: ?|Album(:|/Single) )([^\n:]+)(\n|$)", self.all_media_comments
-        )
-        return match.expand(r"\3").strip() if match else ""
+        """
+        Search for the album name in the following order and return the first match:
+        1. See whether all track names contain {album name} EP or LP
+        2. EP or LP is in the release name, deduce the album name from there
+        3. Check whether the release name has it wrapped in quotes
+        """
+        album_in_tracks = {t.album for t in self._tracks if t.album}
+        if len(album_in_tracks) == 1:
+            return list(album_in_tracks)[0]
+
+        album = self.album_name
+        for pat, text in [
+            (r"(((&|#?\b(?!Double|VA|Various)(\w|[^\w| -])+) )+[EL]P)", album),
+            (r"((['\"])([^'\"]+)\2( VA\d+)*)( |$)", album),
+        ]:
+            m = re.search(pat, text)
+            if m:
+                album = m.group(1).strip()
+                if album[0] in "'\"" and album[-1] in "'\"":
+                    return album.strip("\"'")
+                return album
+        return ""
 
     @cached_property
     def album_name(self) -> str:
-        return self.parsed_album_name or self.original_album_name
+        return self.meta["name"]
 
     @cached_property
     def label(self) -> str:
@@ -137,26 +157,24 @@ class Metaguru(Helpers):
         except KeyError:
             return self.meta["publisher"]["@id"]
 
-    @cached_property
-    def parsed_albumartist(self) -> str:
-        match = re.search(r"Artists?:([^\n]+)", self.all_media_comments)
-        return match.expand(r"\1").strip() if match else ""
-
     @cached_property
     def original_albumartist(self) -> str:
-        return self.meta["byArtist"]["name"]
+        m = re.search(r"Artists?:([^\n]+)", self.all_media_comments)
+        return m.group(1).strip() if m else self.meta["byArtist"]["name"]
 
     @cached_property
     def bandcamp_albumartist(self) -> str:
         """Return the official release albumartist.
         It is correct in half of the cases. In others, we usually find the label name.
         """
-        aartist = self.parsed_albumartist or self.original_albumartist
+        aartist = self.original_albumartist
         if self.label == aartist:
-            aartist = self.parse_track_name(self.album_name).get("artist") or aartist
+            split = self.album_name.split(" - ")
+            if len(split) > 1:
+                aartist = split[0]
 
         aartists = Helpers.split_artists([aartist])
-        remixers_str = " ".join(self.remixers)
+        remixers_str = " ".join(self._tracks.other_artists)
 
         def not_remixer(x: str) -> bool:
             return not any(map(lambda y: y in remixers_str, {x, *x.split(" & ")}))
@@ -165,14 +183,15 @@ class Metaguru(Helpers):
         if (
             len(aartists) == 1
             or len(valid) == len(aartists)
-            and len(self.raw_artists) <= 4
+            and len(self._tracks.artists) <= 4
         ):
-            return aartist
-        return ", ".join(valid)
+            ret = aartist
+        else:
+            ret = ", ".join(valid)
+        return ret
 
     @cached_property
     def image(self) -> str:
-        # TODO: Need to test
         image = self.meta.get("image", "")
         return image[0] if isinstance(image, list) else image
 
@@ -185,7 +204,7 @@ class Metaguru(Helpers):
         """
         rel = self.meta.get("datePublished")
         if rel:
-            return datetime.strptime(re.sub(r" [0-9]{2}:.+", "", rel), "%d %b %Y").date()
+            return datetime.strptime(re.sub(r" \d{2}:.+", "", rel), "%d %b %Y").date()
         return rel
 
     @cached_property
@@ -202,18 +221,39 @@ class Metaguru(Helpers):
     def mediums(self) -> int:
         return self.get_vinyl_count(self.disctitle) if self.media.name == "Vinyl" else 1
 
+    @cached_property
+    def general_catalognum(self) -> str:
+        """Find catalog number in the media-agnostic release metadata and cache it."""
+        cats = [t.catalognum for t in self._tracks if t.catalognum]
+        artists = ordset([self.original_albumartist])
+        if not self._singleton or len(self._tracks.artists) > 1:
+            artists.update(self._tracks.artists)
+            artists.update(self._tracks.other_artists)
+
+        catnum = self.parse_catalognum(
+            album=self.meta["name"],
+            description=self.comments,
+            label=self.label if not self._singleton else "",
+            tracks=tuple(self._tracks.raw_names),
+            artists=tuple(artists),
+        )
+        if len(cats) == len(self._tracks) and len(set(cats)) == 1:
+            return list(cats)[0]
+        return catnum
+
     @property
     def catalognum(self) -> str:
-        artists = [self.parsed_albumartist or self.original_albumartist]
-        if not self._singleton or len(self.raw_artists) > 1:
-            artists.extend(self.raw_artists)
-            artists.extend(self.remixers)
-        return self.parse_catalognum(
-            self.meta["name"],
-            self.disctitle,
-            self.media.description + "\n" + self.comments,
-            self.label,
-            artists,
+        """Find catalog number in the media-specific release metadata or return
+        the cached media-agnostic one.
+        """
+        return (
+            self.parse_catalognum(
+                disctitle=self.disctitle,
+                description=self.media.description,
+                label=self.label if not self._singleton else "",
+                tracks=tuple(self._tracks.raw_names),
+            )
+            or self.general_catalognum
         )
 
     @cached_property
@@ -230,84 +270,13 @@ class Metaguru(Helpers):
             return WORLDWIDE
 
     @cached_property
-    def json_tracks(self) -> List[JSONDict]:
-        try:
-            return self.meta["track"]["itemListElement"]
-        except KeyError:
-            return [{"item": self.meta, "position": 1}]
-
-    @cached_property
-    def json_artists(self) -> List[str]:
-        artists = []
-        for item in map(lambda x: x["item"], self.json_tracks):
-            try:
-                artists.append(item["byArtist"]["name"])
-            except KeyError:
-                continue
-        return artists
-
-    @cached_property
-    def tracks(self) -> List[JSONDict]:
-        """Parse relevant details from the tracks' JSON."""
-        names = self.track_names
-        delim = self.track_delimiter(names)
-        names = self.clean_track_names(names, self.catalognum)
-        tracks = []
-        for item, position in map(op.itemgetter("item", "position"), self.json_tracks):
-            initial_name = names[position - 1]
-            name = self.clear_digi_only(initial_name)
-            track: JSONDict = defaultdict(
-                str,
-                digi_only=name != initial_name,
-                index=position,
-                track_id=item.get("@id"),
-                length=self.get_duration(item) or None,
-                **self.parse_track_name(self.clean_name(name), delim),
-            )
-            lyrics = item.get("recordingOf", {}).get("lyrics", {}).get("text")
-            if lyrics:
-                track["lyrics"] = lyrics.replace("\r", "")
-            tracks.append(track)
-
-        return self.adjust_artists(tracks, self.bandcamp_albumartist)
-
-    @cached_property
-    def remixers(self) -> Set[str]:
-        titles = " ".join(self.track_names)
-        names = re.finditer(r"\( *([^)]+) (?i:(re)?mix|edit)\)", titles, re.I)
-        ft = re.finditer(r"[( ](f(ea)?t[.]? [^()]+)\)?", titles, re.I)
-        return set(map(lambda x: x.expand(r"\1"), it.chain(names, ft)))
-
-    @cached_property
-    def track_artists(self) -> List[str]:
-        return list(filter(op.truth, map(lambda x: x.get("artist") or "", self.tracks)))
+    def tracks(self) -> Tracks:
+        self._tracks.adjust_artists(self.bandcamp_albumartist, self._singleton)
+        return self._tracks
 
     @cached_property
     def unique_artists(self) -> List[str]:
-        return self.split_artists(self.track_artists)
-
-    @cached_property
-    def track_names(self) -> List[str]:
-        raw_tracks = self.meta.get("tracks") or []
-        if raw_tracks:
-            return list(map(lambda x: x.split(". ", maxsplit=1)[1], raw_tracks))
-
-        for item in map(lambda x: x["item"], self.json_tracks):
-            name = item["name"]
-            artist = item.get("byArtist", {}).get("name")
-            if not self._singleton and artist and artist != self.label:
-                name = artist + " - " + name
-            raw_tracks.append(name)
-
-        return raw_tracks
-
-    @cached_property
-    def raw_artists(self) -> List[str]:
-        def only_artist(name: str) -> str:
-            return re.sub(r" - .*", "", PATTERNS["track_alt"].sub("", name))
-
-        artists = list(map(only_artist, filter(lambda x: " - " in x, self.track_names)))
-        return self.split_artists(artists)
+        return self.split_artists(self._tracks.artists)
 
     @cached_property
     def albumartist(self) -> str:
@@ -317,15 +286,17 @@ class Metaguru(Helpers):
         if self.va:
             return self.va_name
 
+        aartist = self.original_albumartist
         if self.unique_artists:
-            return ", ".join(sorted(self.unique_artists))
-        return self.original_albumartist
+            aartist = ", ".join(sorted(self.unique_artists))
+
+        return aartist
 
     @cached_property
     def vinyl_disctitles(self) -> str:
         return " ".join([m.title for m in self.media_formats if m.name == "Vinyl"])
 
-    def search_albumtype(self, word: str) -> bool:
+    def _search_albumtype(self, word: str) -> bool:
         """Return whether the given word (ep or lp) matches the release albumtype.
         True when one of the following conditions is met:
         * if {word}[0-9] is found in the catalognum
@@ -335,27 +306,31 @@ class Metaguru(Helpers):
         """
         sentences = re.split(r"[.]\s+|\n", self.all_media_comments)
         word_pat = re.compile(fr"\b{word}\b", re.I)
-        catnum_pat = re.compile(fr"{word}[0-9]", re.I)
+        catnum_pat = re.compile(fr"{word}\d", re.I)
         name_pat = re.compile(fr"\b(this|{re.escape(self.clean_album_name)})\b", re.I)
         return bool(
             catnum_pat.search(self.catalognum)
-            or word_pat.search(self.original_album_name + " " + self.vinyl_disctitles)
+            or word_pat.search(self.album_name + " " + self.vinyl_disctitles)
             or any(map(lambda s: word_pat.search(s) and name_pat.search(s), sentences))
         )
 
     @cached_property
-    def is_single(self) -> bool:
-        return self._singleton or len(self.track_names) == 1
+    def is_single_album(self) -> bool:
+        return (
+            self._singleton
+            or len({t.main_title for t in self.tracks}) == 1
+            or len(self._tracks.raw_names) == 1
+        )
 
     @cached_property
     def is_lp(self) -> bool:
         """Return whether the release is an LP."""
-        return self.search_albumtype("lp")
+        return self._search_albumtype("lp")
 
     @cached_property
     def is_ep(self) -> bool:
         """Return whether the release is an EP."""
-        return self.search_albumtype("ep")
+        return self._search_albumtype("ep")
 
     def check_albumtype_in_descriptions(self) -> str:
         """Count 'lp', 'album' and 'ep' words in the release and media descriptions
@@ -377,14 +352,16 @@ class Metaguru(Helpers):
         def first_one(artist: str) -> str:
             return PATTERNS["split_artists"].split(artist.replace(" & ", ", "))[0]
 
-        truly_unique = set(map(first_one, self.track_artists))
-        return bool(
-            re.search(r"compilation|best of|anniversary", self.album_name, re.I)
-        ) or (len(truly_unique) > 3 and len(self.tracks) > 4)
+        truly_unique = set(map(first_one, self.tracks.artists))
+        return (
+            bool(re.search(r"compilation|best of|anniversary", self.album_name, re.I))
+            or self._search_albumtype("compilation")
+            or (len(truly_unique) > 3 and len(self.tracks) > 4)
+        )
 
     @cached_property
     def albumtype(self) -> str:
-        if self.is_single:
+        if self._singleton:
             return "single"
         if self.is_ep:
             return "ep"
@@ -403,15 +380,21 @@ class Metaguru(Helpers):
     @cached_property
     def albumtypes(self) -> str:
         albumtypes = {self.albumtype}
-        if self.albumtype == "compilation":
-            albumtypes.add("album")
+        if self.is_comp:
+            if self.albumtype == "ep":
+                albumtypes.add("compilation")
+            else:
+                albumtypes.add("album")
         if self.is_lp:
             albumtypes.add("lp")
-        if len({t["main_title"] for t in self.tracks}) == 1:
+        if self.is_single_album:
             albumtypes.add("single")
-        for word in ["remix", "live", "soundtrack"]:
+        for word in ["remix", "rmx", "edits", "live", "soundtrack"]:
             if word in self.album_name.lower():
-                albumtypes.add(word)
+                albumtypes.add(word.replace("rmx", "remix").replace("edits", "remix"))
+        rmxers = [t.remixer for t in self.tracks if t.remixer and t.remixer != "Original"]
+        if len(rmxers) == len(self.tracks):
+            albumtypes.add("remix")
 
         return "; ".join(sorted(albumtypes))
 
@@ -445,41 +428,26 @@ class Metaguru(Helpers):
         if genre_cfg["maximum"]:
             genres = it.islice(genres, genre_cfg["maximum"])
 
-        return ", ".join(sorted(genres)) or None
+        return ", ".join(sorted(genres)).strip() or None
 
     @cached_property
-    def album_name_with_eplp(self) -> str:
-        match = re.search(r"[:-] ?([A-Z][\w ]+ ((?!an )[EL]P))", self.all_media_comments)
-        return match.expand(r"\1") if match else ""
+    def eplp_album_comments(self) -> str:
+        m = re.search(r"[:-] ?([A-Z][\w ]+ ((?!an )[EL]P))", self.all_media_comments)
+        return m.group(1) if m else ""
 
     @cached_property
     def clean_album_name(self) -> str:
-        if self.parsed_album_name:
-            return self.parsed_album_name
+        if self.official_album_name:
+            return self.official_album_name
 
-        album = self.album_name
-        # look for something in quotes
-        match = re.search(r"(?:^| )(['\"])(.+?)\1(?: |$)", album)
-        if match:
-            album = match.expand(r"\2")
-        album = self.clean_name(album, self.catalognum, remove_extra=True)
-        if (
-            album
-            and not re.search(r"\W | [EL]P", album)  # no delimiters
-            and album not in self.unique_artists  # and it isn't one of the artists
-        ):
-            return album
+        album = self.parsed_album_name or self.album_name
+        artists = [*self.tracks.raw_artists, *self.tracks.artists]
+        artists.sort(key=len, reverse=True)
+        album = self.clean_album(album, self.catalognum, *artists, label=self.label)
 
-        if " EP" in album or " LP" in album:
-            album = self.clean_ep_lp_name(album, self.unique_artists)
-        else:
-            album = self.clean_name(
-                album,
-                self.bandcamp_albumartist,
-                *self.unique_artists,
-                self.parsed_albumartist,
-            )
-        return album or self.album_name_with_eplp or self.catalognum or self.album_name
+        if album.startswith("("):
+            album = self.album_name
+        return album or self.eplp_album_comments or self.catalognum or self.album_name
 
     @property
     def _common(self) -> JSONDict:
@@ -501,13 +469,7 @@ class Metaguru(Helpers):
     @property
     def _common_album(self) -> JSONDict:
         common_data: JSONDict = dict(album=self.clean_album_name)
-        fields = [
-            "label",
-            "catalognum",
-            "albumtype",
-            "albumstatus",
-            "country",
-        ]
+        fields = ["label", "catalognum", "albumtype", "country"]
         if NEW_BEETS:
             fields.extend(["genre", "style", "comments", "albumtypes"])
         common_data.update(self.get_fields(fields))
@@ -517,21 +479,16 @@ class Metaguru(Helpers):
 
         return common_data
 
-    def _trackinfo(self, track: JSONDict, **kwargs: Any) -> TrackInfo:
-        track.pop("digi_only", None)
-        track.pop("main_title", None)
-        ft = track.pop("ft", None)
-        if ft:
-            track["artist"] += f" {ft}"
+    def _trackinfo(self, track: Track, **kwargs: Any) -> TrackInfo:
+        data = track.info
+        data.update(**self._common, **kwargs)
+        if not data["catalognum"] or data["catalognum"] == self.catalognum:
+            data.pop("catalognum", None)
+        if not data["lyrics"]:
+            data.pop("lyrics", None)
         if not NEW_BEETS:
-            track.pop("lyrics", None)
-        track["track_alt"] = track["track_alt"] or None
-        if not track["artist"]:
-            track["artist"] = self.albumartist
-
-        data = dict(**track, **self._common, **kwargs)
-        if "index" in data:
-            data.update(medium_index=data["index"])
+            data.pop("catalognum", None)
+            data.pop("lyrics", None)
         for field in set(data.keys()) & self.excluded_fields:
             data.pop(field)
 
@@ -541,31 +498,28 @@ class Metaguru(Helpers):
     def singleton(self) -> TrackInfo:
         self._singleton = True
         self.media = self.media_formats[0]
-        track_dict = self.tracks[0]
-        if not track_dict["artist"]:
-            track_dict["artist"] = self.bandcamp_albumartist
-        track = self._trackinfo({**track_dict, "index": None})
+        track = list(self.tracks)[0]
+        track = self._trackinfo(track)
         if NEW_BEETS:
             track.update(self._common_album)
             track.pop("album", None)
-            track.pop("albumstatus", None)
         track.track_id = track.data_url
         return track
 
     def _album(self, media: MediaInfo) -> AlbumInfo:
         """Return album for the appropriate release format."""
         self.media = media
-        tracks: Iterable[JSONDict] = self.tracks
         include_digi = self.config.get("include_digital_only_tracks")
-        if not include_digi and self.media.name != DIGI_MEDIA:
-            tracks = it.filterfalse(op.itemgetter("digi_only"), tracks)
 
-        tracks = list(map(op.methodcaller("copy"), tracks))
+        tracks = list(self.tracks)
+        if not include_digi and self.media.name != DIGI_MEDIA:
+            tracks = [t for t in self.tracks if not t.digi_only]
+
         get_trackinfo = partial(
             self._trackinfo,
             medium=1,
             disctitle=self.disctitle or None,
-            medium_total=len(tracks),
+            medium_total=len(self.tracks),
         )
         album_info = AlbumInfo(
             **self._common,
@@ -573,6 +527,7 @@ class Metaguru(Helpers):
             artist=self.albumartist,
             album_id=self.album_id,
             mediums=self.mediums,
+            albumstatus=self.albumstatus,
             tracks=list(map(get_trackinfo, tracks)),
         )
         for key, val in self.get_fields(["va"]).items():
